# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General utilities."""

from lightning_fabric.utilities.apply_func import move_data_to_device
from lightning_fabric.utilities.data import AttributeDict, suggested_max_num_workers
from lightning_fabric.utilities.distributed import is_shared_filesystem
from lightning_fabric.utilities.enums import LightningEnum
from lightning_fabric.utilities.rank_zero import (
    rank_zero_deprecation,
    rank_zero_info,
    rank_zero_only,
    rank_zero_warn,
)
from lightning_fabric.utilities.throughput import Throughput, ThroughputMonitor, measure_flops
from lightning_fabric.utilities.warnings import disable_possible_user_warnings

__all__ = [
    "disable_possible_user_warnings",
    "is_shared_filesystem",
    "LightningEnum",
    "measure_flops",
    "move_data_to_device",
    "rank_zero_deprecation",
    "rank_zero_info",
    "rank_zero_only",
    "rank_zero_warn",
    "suggested_max_num_workers",
    "AttributeDict",
    "Throughput",
    "ThroughputMonitor",
]
